@@ -43,6 +43,12 @@ static int ext2_open_fs(struct btrfs_convert_context *cctx, const char *name)
 			fprintf(stderr, "ext2fs_open: %s\n", error_message(ret));
 		return -1;
 	}
+
+	if (ext2_fs->super->s_feature_incompat & EXT3_FEATURE_INCOMPAT_RECOVER) {
+		error("source filesystem requires recovery, run e2fsck first");
+		goto fail;
+	}
+
 	/*
 	 * We need to know exactly the used space, some RO compat flags like
 	 * BIGALLOC will affect how used space is present.
@@ -86,12 +92,13 @@ static int ext2_open_fs(struct btrfs_convert_context *cctx, const char *name)
 
 	cctx->fs_data = ext2_fs;
 	cctx->blocksize = ext2_fs->blocksize;
-	cctx->block_count = ext2_fs->super->s_blocks_count;
-	cctx->total_bytes = (u64)ext2_fs->super->s_blocks_count * ext2_fs->blocksize;
-	cctx->volume_name = strndup((char *)ext2_fs->super->s_volume_name, 16);
+	cctx->block_count = ext2fs_blocks_count(ext2_fs->super);
+	cctx->total_bytes = cctx->block_count * cctx->blocksize;
+	cctx->label = strndup((char *)ext2_fs->super->s_volume_name, 16);
 	cctx->first_data_block = ext2_fs->super->s_first_data_block;
 	cctx->inodes_count = ext2_fs->super->s_inodes_count;
 	cctx->free_inodes_count = ext2_fs->super->s_free_inodes_count;
+	memcpy(cctx->fs_uuid, ext2_fs->super->s_uuid, SOURCE_FS_UUID_SIZE);
 	return 0;
 fail:
 	ext2fs_close(ext2_fs);
@@ -175,9 +182,9 @@ static int ext2_read_used_space(struct btrfs_convert_context *cctx)
 
 static void ext2_close_fs(struct btrfs_convert_context *cctx)
 {
-	if (cctx->volume_name) {
-		free(cctx->volume_name);
-		cctx->volume_name = NULL;
+	if (cctx->label) {
+		free(cctx->label);
+		cctx->label = NULL;
 	}
 	ext2fs_close(cctx->fs_data);
 	ext2fs_free(cctx->fs_data);
@@ -523,7 +530,7 @@ static int ext2_copy_single_xattr(struct btrfs_trans_handle *trans,
 	strncat(namebuf, EXT2_EXT_ATTR_NAME(entry), entry->e_name_len);
 	if (name_len + datalen > BTRFS_LEAF_DATA_SIZE(root->fs_info) -
 	    sizeof(struct btrfs_item) - sizeof(struct btrfs_dir_item)) {
-		fprintf(stderr, "skip large xattr on inode %Lu name %.*s\n",
+		fprintf(stderr, "skip large xattr on inode %llu name %.*s\n",
 			objectid - INO_OFFSET, name_len, namebuf);
 		goto out;
 	}
@@ -692,7 +699,7 @@ static void ext2_copy_inode_item(struct btrfs_inode_item *dst,
 	memset(&dst->reserved, 0, sizeof(dst->reserved));
 }
 
-#if HAVE_OWN_EXT4_EPOCH_MASK_DEFINE
+#if HAVE_EXT4_EPOCH_MASK_DEFINE
 
 /*
  * Copied and modified from fs/ext4/ext4.h
@@ -763,7 +770,7 @@ out:
 	return ret;
 }
 
-#else /* HAVE_OWN_EXT4_EPOCH_MASK_DEFINE */
+#else /* HAVE_EXT4_EPOCH_MASK_DEFINE */
 
 static int ext4_copy_inode_timespec_extra(struct btrfs_inode_item *dst,
 				ext2_ino_t ext2_ino, u32 s_inode_size,
@@ -780,7 +787,7 @@ static int ext4_copy_inode_timespec_extra(struct btrfs_inode_item *dst,
 	return 0;
 }
 
-#endif /* !HAVE_OWN_EXT4_EPOCH_MASK_DEFINE */
+#endif /* !HAVE_EXT4_EPOCH_MASK_DEFINE */
 
 static int ext2_check_state(struct btrfs_convert_context *cctx)
 {
